/* 
   nget, neon HTTP GET tester
   Copyright (C) 2000, Joe Orton <joe@orton.demon.co.uk>
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   $Id: nget.c,v 1.12 2000/07/16 19:56:52 joe Exp $
*/

/* Would be nice to make this at least as useful as wget for doing 
 * general HTTP downloads. So, TODO:
 *   Proxy support
 *   Long options configurable
 *   Resume downloads
 *   Resume-on-timeout
 * nget wins over wget with:
 *   SSL (don't think wget has this).
 *   Nicer progress bar (IMHO :)
 *   Decent authentication
 */

#include <sys/types.h>

#include <sys/stat.h>
#include <sys/time.h>

#include <stdio.h>
#include <errno.h>
#include <string.h>

#include <neon_config.h>
#include <http_request.h>
#include <http_basic.h>
#include <uri.h>
#include <socket.h>

#include <getopt.h>

#include "basename.h"

static int in_progress = 0, use_stdout = 0;

static void pretty_progress_bar(void *ud, size_t progress, size_t total);
static void conn_notify(void *userdata, sock_status status, const char *info);

static const char *use_filename(struct uri *uri)
{
    const char *fname;

    fname = base_name(uri->path);
    if (strcmp(fname, "/") == 0 || strlen(fname) == 0) {
	fname = "index.html";
    }
    
    return fname;
}

static void parse_args(int argc, char **argv)
{
    if (argc < 2 || argc > 3) {
	printf("nget: Usage: \n"
	       "  `nget url': download url using appropriate filename\n"
	       "  `nget url filename': download url to given filename\n"
	       "  `nget url -': download url and display on stdout\n");
	exit(-1);
    }
}

int main(int argc, char **argv)
{
    http_session *sess;
    struct uri uri = {0}, defaults = {0};
    struct stat st;
    int ret;
    const char *fname;
    FILE *f;

    parse_args(argc, argv);

    sock_register_progress(pretty_progress_bar, NULL);
    sock_register_notify(conn_notify, NULL);

    defaults.port = 0;
    
    if (uri_parse(argv[1], &uri, &defaults) || 
	!uri.path || !uri.host || uri.port == -1) {
	printf("nget: Invalid URL.\n");
	return -1;
    }
    
    if (argc == 3) {
	fname = argv[2];
    } else {
	fname = use_filename(&uri);
    }
    if (strcmp(fname, "-") == 0) {
	f = stdout;
	use_stdout = 1;
    } else if (stat(fname, &st) == 0) {
	printf("nget: File `%s' already exists.\n", fname);
	return -1;
    } else {
	f = fopen(fname, "w");
	if (f == NULL) {
	    printf("nget: Could not open %s: %s\n", fname, strerror(errno));
	    return -1;
	}
    }
    
    neon_debug_init(stderr, 0);

    sock_init();

    sess = http_session_create();
    http_set_useragent(sess, "nget/" NEON_VERSION);

    if (strcasecmp(uri.scheme, "https") == 0) {
	if (uri.port == 0) {
	    uri.port = 443;
	}
	if (http_set_secure(sess, 1) || 
	    http_set_accept_secure_upgrade(sess, 1)) {
	    fprintf(stderr, "nget: SSL not supported.\n");
	    exit(-1);
	}
    }
    
    if (uri.port == 0) {
	uri.port = 80;
    }
    
    http_session_server(sess, uri.host, uri.port);

    if (!use_stdout) {
	printf("nget: Downloading %s to %s\n", argv[1], fname);
    }
    ret = http_get(sess, uri.path, f);

    if (in_progress) {
	printf("\n");
    }

    if (!use_stdout) {
	if (fclose(f)) {
	    fprintf(stderr, "nget: Could not write to disk: %s\n", 
		    strerror(errno));
	}
    }

    switch (ret) {
    case HTTP_OK:
	if (!use_stdout) printf("nget: Download complete.\n");
	ret = 0;
	break;
	/* can put in HTTP_TIMEOUT + restart with ranges support here */
    default:
	fprintf(stderr, 
		"nget: Download failed: %s\n", http_get_error(sess));
	break;
    }


    http_session_destroy(sess);

    return ret;
}

/* Smooth progress bar from cadaver.
 *
 * Doesn't update the bar more than once every 100ms, since this might
 * give flicker, and would be bad if we are displaying on a slow link
 * anyway. */
static void pretty_progress_bar(void *ud, size_t progress, size_t total)
{
    int len, n, update_all = 1;
    double pc;
    static struct timeval last_update = {0};
    static int spin = 0;
    struct timeval this_call;
    if (use_stdout) return;
    in_progress = 1;
    if (total == -1) {
	printf("\rProgress: %d bytes", progress);
	return;
    }
    if (progress < total && gettimeofday(&this_call, NULL) == 0) {
	struct timeval diff;
	timersub(&this_call, &last_update, &diff);
	if (diff.tv_sec == 0 && diff.tv_usec < 20000) {
	    return;
	}
	if (diff.tv_sec == 0 && diff.tv_usec < 100000) {
	    update_all = 0;
	} else {
	    last_update = this_call;
	}
    }
    if (update_all) {
	if (progress == 0 || total == 0) {
	    pc = 0;
	} else {
	    pc = (double)progress / total;
	}
	len = pc * 30;
	printf("\rProgress: [");
	for (n = 0; n<30; n++) {
	    putchar((n<len-1)?'=':
		    (n==(len-1)?'>':' '));
	}
	printf("] %5.1f%% of %d bytes ", pc*100, total);
    } else {
	putchar('\b');
    }   
    putchar(spin==0?'|':((spin==1)?'/':((spin==2)?'-':'\\')));
    spin = (spin+1)%4;
    fflush(stdout);
}

static void conn_notify(void *userdata, sock_status status, const char *info)
{
    switch (status) {
    case sock_namelookup:
    case sock_connecting:
    case sock_connected:
	break;
    case sock_secure_details:
	fprintf(stderr, "nget: Using a secure connection (%s)\n", info);
	break;
    }
}
